﻿-- (c) eradicator a.k.a lossycrypt, 2017-2021, not seperately licensable
-- -------------------------------------------------------------------------- --

--[[ Notes:
  ]]

--[[ Annecdotes:
  ]]

--[[ Future:
  
  If LQ/HQ distinction is ever removed it might be feasible to increase
  animation_speed to make the low-fps-animation look smooth again.

  More agressively downsample rarely used things like explosions?
  
  Mod setting: "Downsample very small animations too"
  
  ]]
  
--[[ Todo:
  ]]
  
-- -------------------------------------------------------------------------- --
-- Built-In                                                                   --
-- -------------------------------------------------------------------------- --
local elroot = (pcall(require,'erlib/empty')) and '' or '__eradicators-library__/'
local say,warn,err,elreq,flag,ercfg=table.unpack(require(elroot..'erlib/shared'))

-- -------------------------------------------------------------------------- --
-- Eradicators Library                                                        --
-- (Factorio does not allow runtime require!)                                 --
-- -------------------------------------------------------------------------- --
-- local log         = elreq('erlib/lua/Log'          )().Logger  'animation-downsampler'
-- local stop        = elreq('erlib/lua/Error'        )().Stopper 'animation-downsampler'
local Table       = elreq('erlib/lua/Table'        )()

local dpairs      = elreq('erlib/lua/Iter/dpairs'  )()

local Setting     = elreq('erlib/factorio/Setting'   )()

-- -------------------------------------------------------------------------- --
-- Constants                                                                  --
-- -------------------------------------------------------------------------- --
local import = PluginManager.make_relative_require 'animation-downsampler'
local const  = import '/const'

-- -------------------------------------------------------------------------- --
-- Module                                                                     --
-- -------------------------------------------------------------------------- --
local anim_frame_keep_factor = {
  lq = 1/100 * Setting.get_value('startup', const.name.setting.lq_vram_percentage),
  hq = 1/100 * Setting.get_value('startup', const.name.setting.hq_vram_percentage),
  }

local min_frames_to_keep = 
  Setting.get_value('startup', const.name.setting.min_frames_to_keep)

-- -------------------------------------------------------------------------- --
-- Local Library                                                              --
-- -------------------------------------------------------------------------- --
  
--[[ Calculates the frame that's "closest" to the target
  
original = {1,2,3,4,5,6,7,8,9,10}
new      = {1,      5,      9   }
nearest  = {1,1,1,5,5,5,5,9,9,9 }
  
new      = {1,    4     7     10}
nearest  = {1,1,4,4,4,7,7,7,10,10}

new      = {   2.5    5    7.5   }
nearest  = {2,2,2,2,5,5,5,7,7,7,7} ???

]]

-- Uses nearest-neighbour downsampling to create
-- smooth frame sequences.
local function downsample_frame_sequence(factor, sequence)
  if factor >= 1 then return sequence end
  if factor == 0 then 
    -- hr/lq versions must have identical "frame_count"
    -- so instead of setting frame_count to 1 this
    -- generates an unanimated sequence instead.
    local n = sequence[math.floor(#sequence/2)] -- nicer explosions
    local r = {}; for i = 1, #sequence do r[i] = n end
    return r end
  --
  -- guarantee min_frames_to_keep
  factor = math.max((min_frames_to_keep - 1) / #sequence, factor)
  --
  local old_frame_count = #sequence
  local virtual_frame_distance = old_frame_count / (old_frame_count * factor)
  --
  local last_virtual = nil
  local next_virtual = 1
  -- local next_virtual = virtual_frame_distance / 2 -- future: better entry point?
  local function update_virtual()
    last_virtual = next_virtual
    next_virtual = math.min(last_virtual + virtual_frame_distance, old_frame_count)
    end
  --
  local function closest(i)
    local a, b = i - last_virtual, next_virtual - i
    return (a <= b) and last_virtual or next_virtual
    end
  --
  local r = {}
  for i = 1, old_frame_count do
    if i >= next_virtual then update_virtual() end
    r[i] = sequence[math.floor(closest(i))]
    end
  --
  return r end


local function is_hr_version(path)
  for i=#path, 1, -1 do if path[i] == 'hr_version' then return true end end
  return false end

local function get_frame_size(tbl)
  local  t = type(tbl.size)
  if     t == 'nil'    then return tbl.width   * tbl.height
  elseif t == 'table'  then return tbl.size[1] * tbl.size[2]
  elseif t == 'number' then return tbl.size    ^ 2
  else error('Invalid animation size type: '..serpent.block(tbl)) end
  end

local function get_effective_sheet_size(tbl)
  return tbl.frame_count * get_frame_size(tbl)
  end
  
local function reduce_frame_count(type)
  for k, v, tbl, path in dpairs(data.raw[type], {include_path = true}) do
    if (k == 'frame_count')
    and (v > min_frames_to_keep)
    and (v <= const.defines.frame_sequence_max_length)
    and (get_effective_sheet_size(tbl)
         > const.defines.min_animation_size_to_optimize)
    then
      local factor = is_hr_version(path)
        and anim_frame_keep_factor.hq
         or anim_frame_keep_factor.lq
      tbl.frame_sequence = downsample_frame_sequence(
        factor, tbl.frame_sequence or Table.range(v)
        )
      end
    end
  end

  
-- -------------------------------------------------------------------------- --
-- Data                                                                       --
-- -------------------------------------------------------------------------- --
  
if (anim_frame_keep_factor.lq < 1)
or (anim_frame_keep_factor.hq < 1) then
  
  -- Some entites like artillery wagon, radar, ...
  -- use RotatedSprite instead of Animation.
  
  -- Some prototypes with high visual impact on the factory are
  -- excluded to not make the game too ugly.
  
  -- Machines
  reduce_frame_count 'assembling-machine'
  reduce_frame_count 'furnace'
  reduce_frame_count 'mining-drill'
  reduce_frame_count 'rocket-silo'
  
  -- Power
  reduce_frame_count 'generator'
  reduce_frame_count 'burner-generator'
  reduce_frame_count 'electric-energy-interface'
  reduce_frame_count 'accumulator'
  
  -- Infrastructure
  reduce_frame_count 'roboport'
  reduce_frame_count 'beacon'
  reduce_frame_count 'gate'
  reduce_frame_count 'lab'
  reduce_frame_count 'logistic-container'
  reduce_frame_count 'infinity-container'
  reduce_frame_count 'pump'
  reduce_frame_count 'offshore-pump'
  reduce_frame_count 'player-port'
  reduce_frame_count 'power-switch'
  reduce_frame_count 'unit-spawner'
  
  -- Turret
  reduce_frame_count 'artillery-turret'
  reduce_frame_count 'fluid-turret'
  reduce_frame_count 'turret'
  
  -- Equipment
  reduce_frame_count 'roboport-equipment'
  
  -- Particles etc
  reduce_frame_count 'explosion'
  reduce_frame_count 'smoke'
  reduce_frame_count 'trivial-smoke'
  reduce_frame_count 'smoke-with-trigger'
  reduce_frame_count 'optimized-particle'
  reduce_frame_count 'fire'
  reduce_frame_count 'sticker'
  reduce_frame_count 'stream'
  reduce_frame_count 'projectile'
  reduce_frame_count 'animation'
  reduce_frame_count 'beam'
  reduce_frame_count 'artillery-flare'
  reduce_frame_count 'rocket-silo-rocket'
  
  -- Units
  reduce_frame_count 'unit'
  reduce_frame_count 'combat-robot'
  reduce_frame_count 'construction-robot'
  reduce_frame_count 'logistic-robot'

  -- Misc
  reduce_frame_count 'simple-entity'
  reduce_frame_count 'simple-entity-with-force'
  reduce_frame_count 'simple-entity-with-owner'
  reduce_frame_count 'resource'
  reduce_frame_count 'corpse'
  reduce_frame_count 'character-corpse'
  -- reduce_frame_count 'car'
  -- reduce_frame_count 'character'
  -- reduce_frame_count 'utility-sprites'
  
  -- Belt
  -- (Low VRAM impact, extremely high visual impact)
  -- reduce_frame_count 'splitter'
  -- reduce_frame_count 'transport-belt'
  -- reduce_frame_count 'underground-belt'
  -- reduce_frame_count 'loader'
  
  -- Train
  -- (Base game already has an option to reduce rotations)
  -- reduce_frame_count 'train-stop'
  -- reduce_frame_count 'rail-chain-signal'
  -- reduce_frame_count 'rail-signal'
  -- reduce_frame_count 'artillery-wagon'
  -- reduce_frame_count 'cargo-wagon'
  -- reduce_frame_count 'fluid-wagon'
  -- reduce_frame_count 'locomotive'
  
  end